#!/usr/bin/env python3
import numpy as np
import networkx as nx
import re
from os.path import isfile
from itertools import combinations
from mopacamk import MOPACamk
from ase.optimize import BFGS
from ase.io import read
from ase.units import kcal, mol
from AMK_parameters import nbonds,emax,method,task,relscf,active,max_val,min_val,startd,label
from createMat import get_G_index

def closestDistanceBetweenLines(a0,a1,b0,b1,clampAll=False,clampA0=False,clampA1=False,clampB0=False,clampB1=False):
    ''' Given two lines defined by numpy.array pairs (a0,a1,b0,b1)
        Return the closest points on each segment and their distance
    '''
    # If clampAll=True, set all clamps to True
    if clampAll: clampA0=True; clampA1=True; clampB0=True; clampB1=True
    # Calculate denomitator
    A = a1 - a0; B = b1 - b0
    magA = np.linalg.norm(A); magB = np.linalg.norm(B)
    _A = A / magA; _B = B / magB
    cross = np.cross(_A, _B);
    denom = np.linalg.norm(cross)**2
    # If lines are parallel (denom=0) test if lines overlap.
    # If they don't overlap then there is a closest point solution.
    # If they do overlap, there are infinite closest positions, but there is a closest distance
    if not denom:
        d0 = np.dot(_A,(b0-a0))
        # Overlap only possible with clamping
        if clampA0 or clampA1 or clampB0 or clampB1:
            d1 = np.dot(_A,(b1-a0))
            # Is segment B before A?
            if d0 <= 0 >= d1:
                if clampA0 and clampB1:
                    if np.absolute(d0) < np.absolute(d1): return a0,b0,np.linalg.norm(a0-b0)
                    return a0,b1,np.linalg.norm(a0-b1)
            # Is segment B after A?
            elif d0 >= magA <= d1:
                if clampA1 and clampB0:
                    if np.absolute(d0) < np.absolute(d1): return a1,b0,np.linalg.norm(a1-b0)
                    return a1,b1,np.linalg.norm(a1-b1)
        # Segments overlap, return distance between parallel segments
        return None,None,np.linalg.norm(((d0*_A)+a0)-b0)
    # Lines criss-cross: Calculate the projected closest points
    t = (b0 - a0);
    detA = np.linalg.det([t, _B, cross])
    detB = np.linalg.det([t, _A, cross])
    t0 = detA/denom; t1 = detB/denom
    pA = a0 + (_A * t0) # Projected closest point on segment A
    pB = b0 + (_B * t1) # Projected closest point on segment B
    # Clamp projections
    if clampA0 or clampA1 or clampB0 or clampB1:
        if clampA0 and t0 < 0: pA = a0
        elif clampA1 and t0 > magA: pA = a1
        if clampB0 and t1 < 0: pB = b0
        elif clampB1 and t1 > magB: pB = b1
        # Clamp projection A
        if (clampA0 and t0 < 0) or (clampA1 and t0 > magA):
            dot = np.dot(_B,(pA-b0))
            if clampB0 and dot < 0: dot = 0
            elif clampB1 and dot > magB: dot = magB
            pB = b0 + (_B * dot)
        # Clamp projection B
        if (clampB0 and t1 < 0) or (clampB1 and t1 > magB):
            dot = np.dot(_A,(pB-a0))
            if clampA0 and dot < 0: dot = 0
            elif clampA1 and dot > magA: dot = magA
            pA = a0 + (_A * dot)
    return np.linalg.norm(pA-pB),pA,pB

def CheckMove(G,bre,cre,ibond):
#check valency
    for k in active:
        if G.degree(k) not in range(min_val[symb[k]],max_val[symb[k]]+1): return False
#check bond formation is possible--> initial distance < startd angs and path does not cross an existing bond
    if len(cre) >= 1:
        for index in cre:
            if d[ind[index]][jnd[index]] > startd: return False
            a0 = np.array(posit[ind[index]])
            a1 = np.array(posit[jnd[index]])
            for i in range(len(ibn)):
                b0 = np.array(posit[ibn[i]])
                b1 = np.array(posit[jbn[i]])
                d34ref = np.linalg.norm(b0 - b1)
                dmin,pA,pB = closestDistanceBetweenLines(a0,a1,b0,b1)
                d1 = np.linalg.norm(pA - a0)
                d2 = np.linalg.norm(pA - a1)
                d3 = np.linalg.norm(pB - b0)
                d4 = np.linalg.norm(pB - b1)
                dm = min(d1,d2,d3,d4)
                d34 = d3 + d4
                diff = (abs(d34 - d34ref) / d34ref) * 100
#Check if the intersect crosses a bond
                if dm > 0.1 and diff < 10: return False
#Check dissociations of a bond with a bo greater than 1.5
    if len(bre) >= 1:
        for index in bre:
            if bo[index] > 1.5: return False
#Check dissociations leading to >=2 atoms or 2 or more bond breakages at the same time
    nd = 0
    for i in range(natom):
        if G.degree(i) == 0: nd += 1
    if nd >= 2 or ibond == len(bre) >= 2: return False

    return True

'''
Program to obtain all combinations of bonds that can be formed/broken
within a number of active atoms and with a max of nbonds involved
active: Active atoms 
emax:   Upper limit for the energy of products
method: Any of the semiempirical methods of MOPAC2016 
nbonds: Max number of bonds involved
startd: Initial max dist between a couple of atoms to make a bond
'''
with open('amk.dat') as inp:
    lines = inp.readlines()
    for line in lines:
        if line.find('active')    !=-1: active = [int(x) - 1 for x in re.sub('[^0-9]',' ', line).split(' ') if x != '' ]
        if line.find('emax')      !=-1: emax   = float(line.split()[1])
        if line.find('molecule')  !=-1: molecu = str(line.split()[1])
        if line.find('method')    !=-1: method = str(line.split()[1])
        if line.find('nbonds')    !=-1: nbonds = float(line.split()[1]) 
        if line.find('startd')    !=-1: startd = float(line.split()[1]) 
        if line.find('neighbors') !=-1: 
            atom = str(line.split()[1]) 
            vale = [int(x) for x in re.sub('[^0-9]',' ', line).split(' ') if x != '' ]
            min_val[atom] = vale[0]; max_val[atom] = vale[1]

#opt_start is the optimized initial structure
rmol   = read(molecu+'.xyz')
natom  = len(rmol)
if len(active) == 0: active = [item for item in range(natom)]

#Setting up initial stuff
symb   = rmol.get_chemical_symbols()
aton   = rmol.get_atomic_numbers()
d      = rmol.get_all_distances(mic=False, vector=False)
posit  = rmol.get_positions()
n_dist = int( natom * (natom - 1) / 2)

#Checking that all neighbors are correctly set
stfile  = open('tsdirLL_'+molecu+'/extforce_stats.log', 'w')
for i in active:
    if symb[i] not in list(min_val.keys()): 
        stfile.write('Define neighbors for element: '+symb[i]+'\n')
        print(symb[i])    
        exit()

#rmol.calc is an instance of the mopac calculation
#Determine initial energy e0 and bond order matrix bo
rmol.calc = MOPACamk(label=label, method=method, relscf=relscf, task=task)
opt = BFGS(rmol,logfile='bfgs.log')
opt.run(fmax=1, steps=50)
e0     = rmol.calc.get_final_heat_of_formation() * mol / kcal
bo     = [float(item) for item in rmol.calc.get_bond_order()]

#Obtaining initial Adjacency and indexes 
act_dist = []
G,ind,jnd,ibn,jbn = get_G_index(rmol,1,natom,False)
for i in range(n_dist):
    if G[ind[i]][jnd[i]]['weight'] == 0: G.remove_edge(ind[i],jnd[i])
    if ind[i] in active and jnd[i] in active: act_dist.append(i)

G_orig = G.copy()
A  = nx.adjacency_matrix(G) ; Ar = A.A
#Initialize variables and dicts 
nts = 0; new = 0
dict_ts = {}; dict_c = {}

#Output files
#tsfile  = 'tsdirLL_'+molecu+'/dict_ts.inp'
tsfile2 = open('tsdirLL_'+molecu+'/ts_bonds.inp', 'w')

#if isfile(tsfile):
#    with open(tsfile) as f:
#        lines = f.readlines()
#        for i, line in enumerate(lines):
#            dict_ts[i+1]=list(np.float_(line.split())) 
#    nts = i + 1

#Print out statistics info
totalnc = 0
for ibnd in range(nbonds):
    lista = list(combinations(act_dist, ibnd+1))
    totalnc += len(lista)
    cadena = 'Number of combinations for '+str(ibnd + 1)+' bond(s): '+str(len(lista))+'\n'
    stfile.write(cadena)
stfile.write('Total number of combinations: '+str(totalnc)+'\n')

nc = 0
#Generate all possible combinations for a max of nbonds 
for ibnd in range(nbonds):
    for comb in list(combinations(act_dist, ibnd+1)):
        nc += 1
        stfile.write(str(nc)+' out of '+str(totalnc)+'\n')
        bre = []; cre = []
        G = G_orig.copy()
        for i in comb:
#Changing G according to combination i
            if G.has_edge(ind[i],jnd[i]):
                G.remove_edge(ind[i],jnd[i])
                bre.append(i)
            else:
                G.add_edge(ind[i],jnd[i])
                cre.append(i)
#Check valencies and stuff
        if CheckMove(G,bre,cre,ibnd+1):
#new adjacency matrices Ap and Ats (employed as ts label)
            A   = nx.adjacency_matrix(G) ; Ap  = A.A
            Ats = 0.5 * (Ar + Ap)
            for i in range(natom): Ats[i][i] = aton[i]
            tag = sorted( [round(elem,3) for elem in np.linalg.eigvals(Ats) ] )
#check if the ts is not repeated 
            if tag not in dict_ts.values():
###Before including a channel, check the energy of products
                eprod = 0
                if not nx.is_connected(G):
                    for c in nx.connected_components(G):
                        pmol = rmol.copy()
                        del pmol[[atom.index for atom in pmol if not atom.index in G.subgraph(c).nodes()]]
                        pmol.calc = MOPACamk(label=label, method=method, relscf=relscf)
                        opt = BFGS(pmol,logfile='bfgs.log')
                        opt.run(fmax=1, steps=50)
                        eprod += pmol.calc.get_final_heat_of_formation() * mol / kcal
                    eprod = eprod - e0 
                if eprod < emax:
                    nts += 1 ; new += 1
                    dict_ts[str(nts)] = tag ; dict_c[str(nts)]  = comb
G = G_orig.copy()            
#Printing ts unique labels 
#with open(tsfile,"w+") as f: 
#    for line in np.matrix([item for item in dict_ts.values()]): np.savetxt(f, line, fmt='%s')
#printing the bonds that change in any one of them
print(new)
for nchan,x in dict_c.items():
    bfbonds=''
    for item in x:
        if G.has_edge(ind[item],jnd[item]): bfbonds=bfbonds+' b '+str(ind[item]+1)+' '+str(jnd[item]+1)
        else:                               bfbonds=bfbonds+' f '+str(ind[item]+1)+' '+str(jnd[item]+1)
    tsfile2.write(bfbonds+'\n')
